(*
   This file is part of Anoq's implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003 Anoq of the Sun (alias Johnny Andersen)

   E-mail: anoq at HardcoreProcessing dot com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information. E-mail:
   info at HardcoreProcessing dot com

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiAbs =
  struct
    type Label = string
    type TyName = string
    type CName = string
    type TyVarName = string
    type MetaVarName = string
    type VId = string

    (* Restrictions for type variables *)
    datatype RestrTyPat =
      RstrTyPat of {args : {name : TyVarName, pat : RestrTyPat} list,
                    name : TyName}

    datatype Restr =
      RstrNone
    | RstrExclude of {labs : Label list}
    | RstrTyPats of {pats : RestrTyPat list}
    
    (* Type system using restrictions *)
    datatype 'a Ty =
      TyCon of {args : {name : TyVarName, ty : 'a Ty} list,
                name : TyName, tags : CName list, ann : 'a}
    | TyFun of {from : 'a Ty, to : 'a Ty, ann : 'a}
    | TyField of {present : 'a Ty, ty : 'a Ty, ann : 'a}
    | TyExtRecord of {fields : {lab : Label, ty : 'a Ty} list,
                      ext : 'a Ty, ann : 'a}
    | TyEmptyRecord of {ann : 'a}
    | TyVar of {name : TyVarName, restr : Restr, ann : 'a} 
    | TyMeta of {name : MetaVarName, restr : Restr,
                 bound : 'a Ty option ref, ann : 'a}
    
    type 'a TyScheme =
      {params : {name : TyVarName, restr : Restr} list,
       ty : 'a Ty, ann : 'a}

    (* Utility type for type inference *)
    type 'a OrderRow =
      {lab : Label, ty : 'a Ty} list

    (* The base constants (These are really part of the supported basis) *)
    datatype SCon =
      SCReal of real
    | SCInt of int
    | SCWord of word
    | SCString of string
    | SCChar of char

    (* The expression langauge *)
    (* Patterns *)
    datatype 'a Pat =
      PatVar of {name : VId, ann : 'a}
    | PatTyped of {pat : 'a Pat, ty : 'a Ty, ann : 'a}
    | PatLayered of {vid : VId, pat : 'a Pat, ann : 'a}
    | PatCon of {tag : CName, tyScheme : 'a TyScheme, ann : 'a}
    | PatParamCon of {tag : CName, tyScheme : 'a TyScheme,
                      pat : 'a Pat, ann : 'a}
    | PatEx of {tag : CName, ann : 'a}
    | PatParamEx of {tag : CName, pat : 'a Pat, ty : 'a Ty, ann : 'a}
    | PatSCon of {scon : SCon, ann : 'a}
    | PatWild of {ann : 'a}
    | PatEmptyRecord of {ann : 'a}
    | PatExtRecord of {fields : {lab : Label, pat : 'a Pat} list,
                       ext : 'a Pat, ann : 'a}
    | PatOptField of {present : 'a Pat, pat : 'a Pat, ann : 'a}

    (* Expressions *)
    datatype 'a Exp =
      ExpFun of {param : VId, exp : 'a Exp, ann : 'a}
    | ExpApp of {funExp : 'a Exp, argExp : 'a Exp, ann : 'a}
    | ExpVar of {name : VId, ann : 'a}
    | ExpTyped of {exp : 'a Exp, ty : 'a Ty, ann : 'a}
    | ExpCon of {tag : CName, tyScheme : 'a TyScheme, ann : 'a}
    | ExpEx of {tag : CName, ty : 'a Ty, ann : 'a}
    | ExpSCon of {scon : SCon, ann : 'a}
    | ExpEmptyRecord of {ann : 'a}
    | ExpExtRecord of {fields : {lab : Label, exp : 'a Exp} list,
                       ext : 'a Exp, ann : 'a}
    | ExpOptField of {present : 'a Exp, exp : 'a Exp, ann : 'a}
    | ExpLet of {binds : {pat : 'a Pat, exp : 'a Exp, ann : 'a} list,
                 exp : 'a Exp, ann : 'a}
    | ExpLetRec of {binds : {pat : 'a Pat, exp : 'a Exp, ann : 'a} list,
                    exp : 'a Exp, ann : 'a}
    | ExpLetEx of {binds : {tag : CName, ty : 'a Ty, ann : 'a} list,
                   exp : 'a Exp, ann : 'a}
    | ExpHandle of {exp : 'a Exp,
                    matches : {pat : 'a Pat, exp : 'a Exp, ann : 'a} list,
                    ann : 'a}
    | ExpRaise of {exp : 'a Exp, ann : 'a}
    | ExpCase of {exp : 'a Exp,
                  matches : {pat : 'a Pat, exp : 'a Exp, ann : 'a} list,
                  ann : 'a}
    | ExpFieldCase of {exp : 'a Exp, tyvar : TyVarName,
                       absExp : 'a Exp,
                       prePat : 'a Pat, preExp : 'a Exp,
                       tyOut : 'a Ty, ann : 'a}
  end
