(*
   This file is part of Anoq's implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003 Anoq of the Sun (alias Johnny Andersen)

   E-mail: anoq at HardcoreProcessing dot com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information. E-mail:
   info at HardcoreProcessing dot com

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiInterpret =
  struct
    exception InterpError

    structure ST = SymbolTable

    (* Some tags which have a special meaning in the semantics *)
    val tagRef = "ref"
    val exMatch = "Match"
    
    fun matchEqConst (KsiAbs.SCInt i1,
                      KsiAbs.SCInt i2) =
          i1 = i2
      | matchEqConst (KsiAbs.SCString s1,
                      KsiAbs.SCString s2) =
          s1 = s2
      | matchEqConst (KsiAbs.SCChar c1,
                      KsiAbs.SCChar c2) =
          c1 = c2
      | matchEqConst (KsiAbs.SCWord w1,
                      KsiAbs.SCWord w2) =
          w1 = w2
      | matchEqConst _ =
          (* The types does not match or we are trying to
             compare two reals - all of which would be an error! *)
          raise InterpError

    (* In the dynamic semantics of ksi-Calculus this is the judgement:
       r \vdash^# lab \Rightarrow v; r' *)
    fun extractLab _ nil acc =
          (* Notice that acc will be reversed here -
             but order is irrelevant for records *)
          (KsiVals.Field NONE, acc)
      | extractLab label ((field as {lab, value})::fields) acc =
          if label = lab then
            (* Notice that acc will be reversed here -
               but order is irrelevant for records *)
            (KsiVals.Field (SOME value), acc @ fields)
          else
            extractLab label fields (field::acc)

    (* In the dynamic semantics of ksi-Calculus this is the
       record rules in the judgement for patterns. *)
    fun evalPatExtRecord env valueFields ((field as {lab, pat})::fields) ext accEnv =
          let
            val (value, restFields) = extractLab lab valueFields nil
          in
            case evalPat env value pat of
              SOME patEnv =>
                evalPatExtRecord env restFields fields ext
                                 (ST.append (patEnv, accEnv))
            | NONE =>
                (* A pattern-match failed - stop evaluating and FAIL *)
                NONE
          end
      | evalPatExtRecord env valueFields nil ext accEnv =
          (case evalPat env (KsiVals.Record valueFields) ext of
             SOME patEnv =>
               SOME (ST.append (patEnv, accEnv))
           | NONE =>
               (* The last pattern-match failed - stop evaluating and FAIL *)
               NONE
          )

    (* In the dynamic semantics of ksi-Calculus this is the judgement:
       \Gamma; v \vdash p \Rightarrow \Gamma' / FAIL *)
    and evalPat env value (KsiAbs.PatVar {name, ann}) =
          SOME (ST.insert (ST.empty ()) (ST.symbol name) value)
      | evalPat env value (KsiAbs.PatTyped {pat, ty, ann}) =
          evalPat env value pat
      | evalPat env value (KsiAbs.PatLayered {vid, pat, ann}) =
          (case evalPat env value pat of
             SOME patEnv =>
               SOME (ST.insert patEnv (ST.symbol vid) value)
           | NONE =>
               NONE
          )
      | evalPat env value (KsiAbs.PatCon {tag, tyScheme, ann}) =
          (case value of
             KsiVals.Constructor {name, value = NONE} =>
               if tag = name then
                 SOME (ST.empty ())
               else
                 NONE
           | _ =>
               (* The constructor either has a value, which is an error here,
                  or the value is not a constructor at all! *)
               raise InterpError
          )
      | evalPat env value (KsiAbs.PatParamCon {tag, tyScheme, pat, ann}) =
          if tag = tagRef then
            (case value of
               KsiVals.Ref (ref v) =>
                 evalPat env v pat
             | _ =>
                 (* The value is not a reference as expected! *)
                 (print ("KsiInterpret.evalPat error: ref pattern " ^
                         "is not matched to a reference value as expected");
                  raise InterpError)
            )
          else
            (case value of
               KsiVals.Constructor {name, value = innerValue} =>
                 if tag = name then
                   (case innerValue of
                      SOME v =>
                        evalPat env v pat
                    | NONE =>
                        (* The constructor does not have a
                           value which is an error here! *)
                        (print ("KsiInterpret.evalPat error: " ^
                                "Constructor pattern with tag " ^ tag ^
                                " does not have a value as expected");
                         raise InterpError)
                   )
                 else
                   NONE
             | _ =>
                 (* The value is not a constructor as expected! *)
                 (print ("KsiInterpret.evalPat error: " ^
                         "Constructor pattern with tag " ^ tag ^
                         " is not matched to a constructor value as expected");
                  raise InterpError)
            )
      | evalPat env value (KsiAbs.PatEx {tag, ann}) =
          (case value of
             KsiVals.Exception {name, value = NONE} =>
               (case ST.lookup env (ST.symbol tag) of
                  SOME en =>
                    (case en of
                       KsiVals.Exception {name = tagName, value = _} =>
                         if tagName = name then
                           SOME (ST.empty ())
                         else
                           NONE
                     | _ =>
                         (* Exception tag not bound to exception name *)
                         (print ("KsiInterpret.evalExp error: Exception tag " ^
                                 tag ^ " not bound to exception name");
                          raise InterpError)
                    )
                | NONE =>
                    (* Unbound exception constructor *)
                    (print ("KsiInterpret.evalExp error: " ^
                            "Unbound exception constructor: " ^ tag);
                     raise InterpError)
               )
           | _ =>
               (* The exception either has a value, which is an error here
                  or it is not an exception value at all! *)
               raise InterpError
          )
      | evalPat env value (KsiAbs.PatParamEx {tag, pat, ty, ann}) =
          (case value of
             KsiVals.Exception {name, value = innerValue} =>
               (case ST.lookup env (ST.symbol tag) of
                  SOME en =>
                    (case en of
                       KsiVals.Exception {name = tagName, value = _} =>
                         if tagName = name then
                           (case innerValue of
                              SOME v =>
                                evalPat env v pat
                            | NONE =>
                                (* The exception does not have a
                                value which is an error here! *)
                                (print ("KsiInterpret.evalPat error: " ^
                                        "Exception pattern with tag " ^ tag ^
                                        " does not have a value as expected");
                                 raise InterpError)
                           )
                         else
                           NONE
                     | _ =>
                         (* Exception tag not bound to exception name *)
                         (print ("KsiInterpret.evalExp error: Exception tag " ^
                                 tag ^ " not bound to an exception name");
                          raise InterpError)
                    )
                | NONE =>
                    (* Unbound exception constructor *)
                    (print ("KsiInterpret.evalExp error: " ^
                            "Unbound exception constructor: " ^ tag);
                     raise InterpError)
               )
           | _ =>
               (* The value is not an exception as expected! *)
                 (print ("KsiInterpret.evalPat error: " ^
                         "Exception pattern with tag " ^ tag ^
                         " is not matched to an exception value as expected");
                  raise InterpError)
          )
      | evalPat env value (KsiAbs.PatSCon {scon, ann}) =
          (case value of
             KsiVals.Const const =>
               if matchEqConst (scon, const) then
                 SOME (ST.empty ())
               else
                 NONE
           | _ =>
               (* Value is not a constant as expected! *)
               (print ("KsiInterpret.evalPat error: " ^
                       "Value is not a constant as expected");
                raise InterpError)
          )
      | evalPat env value (KsiAbs.PatWild {ann}) =
          SOME (ST.empty ())
      | evalPat env value (KsiAbs.PatEmptyRecord {ann}) =
          (case value of
             KsiVals.Record fields =>
               (* We raise an exception if fields is not nil *)
               (case fields of
                  nil =>
                    SOME (ST.empty ())
                | _ =>
                    (print ("KsiInterpret.evalPat error: " ^
                            "The empty record was matched to a non-empty " ^
                            "record value");
                     raise InterpError)
               )
           | _ =>
               (* Value is not a record as expected! *)
               raise InterpError
          )
      | evalPat env value (KsiAbs.PatExtRecord {fields, ext, ann}) =
          (case value of
             KsiVals.Record valueFields =>
               evalPatExtRecord env valueFields fields ext (ST.empty ())
           | _ =>
               (* Value is not a record as expected! *)
               raise InterpError
          )
      | evalPat env value (KsiAbs.PatOptField {present, pat, ann}) =
          (case present of
             KsiAbs.PatCon {tag = "present", tyScheme, ann} =>
               (case value of
                  KsiVals.Field (SOME innerValue) =>
                    (case evalPat env innerValue pat of
                       SOME patEnv =>
                         SOME patEnv
                     | NONE =>
                         (* We require this pattern to be exhaustive,
                            so this is an error! *)
                         raise InterpError
                    )
                | _ =>
                    (* Value not a present optional field as expected! *)
                    raise InterpError
               )
           | _ =>
               (* Constructor not the present tag as required! *)
               raise InterpError
          )

    (* The following is the datatype and the functions
       for handling mutually recursive function bindings. *)
    datatype 'a FunClosure =
      FunClosure of
        {param : KsiAbs.VId,
         exp : 'a KsiAbs.Exp,
         env : KsiVals.T ST.T,
         recEnv : {pat : 'a KsiAbs.Pat, clos : 'a FunClosure} list}

    fun makeFunClosure env recEnv param exp =
          FunClosure
            {param = param,
             exp = exp,
             env = env,
             recEnv = recEnv}

    fun recEnvValue newRecEnv {pat, clos = (FunClosure {param, exp, env, recEnv})} =
          {pat = pat,
           clos = FunClosure
                    {param = param,
                     exp = exp,
                     env = env,
                     recEnv = newRecEnv}
          }

    fun makeRecEnv env =
          map (recEnvValue env) env

    fun insertRecEnvF ({pat, clos as (FunClosure {param, ...})}, env) =
          (case evalPat env (makeEvalFun clos) pat of
             SOME patEnv =>
               ST.append (patEnv, env)
           | NONE =>
               (* Pattern-match failed - but we require let-binds
                  to have exhaustive patterns, so this is an error. *)
               raise InterpError
          )
          
    and insertRecEnv env recEnv =
          foldl insertRecEnvF env recEnv
    
    and makeEvalFun (clos as (FunClosure {param, exp, env, recEnv})) =
          let
            fun f paramValue =
                  evalExp (ST.insert (insertRecEnv env (makeRecEnv recEnv))
                                     (ST.symbol param) paramValue)
                          exp
          in
            KsiVals.Fun (f : KsiVals.T -> KsiVals.T)
          end
    (* End of handling mutually recursive function bindings *)
          
    (* In the dynamic semantics of ksi-Calculus this is the judgement:
       \Gamma \vdash e \Rightarrow v *)
    and evalExp env (KsiAbs.ExpFun {param, exp, ann}) =
          makeEvalFun (makeFunClosure env nil param exp)
      | evalExp env (KsiAbs.ExpApp {funExp, argExp, ann}) =
          (case evalExp env funExp of
             KsiVals.Fun f =>
               (* This does not care about the function closure stuff in
                  the dynamic semantics of xi-Calculus.
                  The f here could even be an ICeXL function,
                  a compiled CeXL function or a predefined built-in
                  function - we just call it :-)
                  Actually we let the assignment operator :=
                  be just a built-in function which can be called here. *)
               f (evalExp env argExp)
           | KsiVals.Constructor {name, value} =>
               if name = tagRef then
                 KsiVals.Ref (ref (evalExp env argExp))
               else
                 KsiVals.Constructor
                   {name = name,
                    value = SOME (evalExp env argExp)}
           | KsiVals.Exception {name, value} =>
               KsiVals.Exception
                 {name = name,
                  value = SOME (evalExp env argExp)}
           | _ =>
               (* Function expression neither evaluated to
                  a function nor a constructor *)
               raise InterpError
          )
      | evalExp env (KsiAbs.ExpVar {name, ann : 'a}) =
          (case ST.lookup env (ST.symbol name) of
             SOME value =>
               value
           | NONE =>
               (* Unbound variable *)
               (print ("KsiInterpret.evalExp error: Unbound variable: " ^ name);
                raise InterpError)
          )
      | evalExp env (KsiAbs.ExpTyped {exp, ty, ann}) =
          evalExp env exp
      | evalExp env (KsiAbs.ExpCon {tag, tyScheme, ann}) =
          KsiVals.Constructor {name = tag, value = NONE}
      | evalExp env (KsiAbs.ExpEx {tag, ty, ann}) =
          (case ST.lookup env (ST.symbol tag) of
             SOME value =>
               (case value of
                  KsiVals.Exception {name, value = _} =>
                    (* We found the exception name in the environment *)
                    KsiVals.Exception {name = name, value = NONE}
                | _ =>
                    (* Environment has a value which is not an exception name as expected *)
                    (print ("KsiInterpret.evalExp error: " ^
                            "The environment has a value which is " ^
                            "not an exception name for the exception tag: " ^
                            tag);
                     raise InterpError)
               )
           | NONE =>
               (* Unbound exception tag *)
               (print ("KsiInterpret.evalExp error: Exception tag: " ^ tag);
                raise InterpError)
          )
      | evalExp env (KsiAbs.ExpSCon {scon, ann}) =
          KsiVals.Const scon
      | evalExp env (KsiAbs.ExpEmptyRecord {ann}) =
          KsiVals.Record nil
      | evalExp env (KsiAbs.ExpExtRecord {fields, ext, ann}) =
          let
            val r = evalExpFields env fields
          in
            case evalExp env ext of
              KsiVals.Record extR =>
                (* The order of evaluation is important so
                   we really have to use list append here. *)
                KsiVals.Record (r @ extR)
            | _ =>
                (* ext did not evaluate to a record! *)
                raise InterpError
          end
      | evalExp env (KsiAbs.ExpOptField {present, exp, ann}) =
          (case evalExp env present of
             KsiVals.Constructor {name = "present", value = NONE} =>
               KsiVals.Field (SOME (evalExp env exp))
           | KsiVals.Constructor {name = "absent", value = NONE} =>
               (* The semantics dictate that we must evaluate
                  exp even though we don't use it.
                  This is for possible side effects. *)
               (evalExp env exp;
                KsiVals.Field NONE)
           | _ =>
               (* Something went completely wrong here! *)
               raise InterpError
          )
      | evalExp env (KsiAbs.ExpLet {binds, exp, ann}) =
          evalExp (evalLetBinds env binds) exp
      | evalExp env (KsiAbs.ExpLetRec {binds, exp, ann}) =
          evalExp (evalLetRecBinds env binds) exp
      | evalExp env (KsiAbs.ExpCase {exp, matches, ann}) =
          evalExpMatches env (evalExp env exp) matches
      | evalExp env (KsiAbs.ExpFieldCase
                       {exp, tyvar, absExp, prePat, preExp,
                        tyOut, ann}) =
          (case evalExp env exp of
             KsiVals.Field (SOME fieldValue) =>
               (case evalPat env fieldValue prePat of
                  SOME patEnv =>
                    evalExp (ST.append (patEnv, env)) preExp
                | NONE =>
                    (* We require that field-case patterns are exhaustive,
                       so this is an error! *)
                    raise InterpError
               )
           | KsiVals.Field NONE =>
               evalExp env absExp
           | _ =>
               (* Value not an optional field as expected! *)
               raise InterpError
           
          )
      | evalExp env (KsiAbs.ExpLetEx {binds, exp, ann}) =
          evalExp (evalExBinds env binds) exp
      | evalExp env (KsiAbs.ExpHandle {matches, exp, ann}) =
          ((evalExp env exp)
              handle
                KsiVals.Pack ex => evalExpExMatches env ex matches
          )
      | evalExp env (KsiAbs.ExpRaise {exp, ann}) =
          raise KsiVals.Pack (evalExp env exp)
    
    and evalExpFieldsF env ({lab, exp}, r) =
          (case evalExp env exp of
             KsiVals.Field (SOME value) =>
               {lab = lab, value = value} :: r
           | KsiVals.Field NONE =>
               r
           | _ =>
               (* The field expression did not evaluate to a field! *)
               raise InterpError
          )

    and evalExpFields env fields =
          (* The list of record fields gets reversed here,
             but the order is irrelevant for records. *)
          foldl (evalExpFieldsF env) nil fields

    and evalLetBindsF env ({pat, exp, ann}, accEnv) =
          (case evalPat env (evalExp env exp) pat of
             SOME patEnv =>
               ST.append (patEnv, accEnv)
           | NONE =>
               (* Pattern-match failed - but we require let-binds
                  to have exhaustive patterns, so this is an error. *)
               raise InterpError
          )

    and evalLetBinds env binds =
          foldl (evalLetBindsF env) env binds

    and evalLetRecBindMakeRecEnv1 env
            {pat, exp = KsiAbs.ExpFun {param, exp, ann = _}, ann} =
          {pat = pat, clos = makeFunClosure env nil param exp}
      | evalLetRecBindMakeRecEnv1 _ _ =
          (* Expression not of the form ExpFun as required! *)
          raise InterpError

    and evalLetRecBindMakePatEnv env ({pat, clos}, accEnv) =
          (case evalPat env (makeEvalFun clos) pat of
             SOME patEnv =>
               ST.append (patEnv, accEnv)
           | NONE =>
               (* Pattern-match failed - but we require let-binds
                  to have exhaustive patterns, so this is an error. *)
               raise InterpError
          )

    and evalLetRecBinds env binds =
          let
            val recEnv1 = map (evalLetRecBindMakeRecEnv1 env) binds
            val recEnv2 = map (recEnvValue recEnv1) recEnv1
          in
            foldl (evalLetRecBindMakePatEnv env) env recEnv2
          end

    and evalExpMatchesF env value ({pat, exp, ann}, NONE) =
          (case evalPat env value pat of
             SOME patEnv =>
               SOME (evalExp (ST.append (patEnv, env)) exp)
           | NONE =>
               NONE
          )
      | evalExpMatchesF _ _ (_, acc as (SOME _)) =
          acc

    and evalExpMatches env value matches =
          (case foldl (evalExpMatchesF env value) NONE matches of
             SOME resValue =>
               resValue
           | NONE =>
               (* Match-error - we raise a Match exception. *)
               (case ST.lookup env (ST.symbol exMatch) of
                  SOME en =>
                    (case en of
                       KsiVals.Exception {name = matchName, value = _} =>
                         raise KsiVals.Pack
                                 (KsiVals.Exception
                                    {name = matchName, value = NONE})
                     | _ =>
                         (* Exception tag not bound to exception name *)
                         (print ("KsiInterpret.evalExp error: " ^
                                 "Exception tag Match not bound to an exception name");
                          raise InterpError)
                    )
                | NONE =>
                    (* Match exception not defined! *)
                    (print ("KsiInterpret.evalExp error: " ^
                            "The exception constructor Match is not defined as expected");
                     raise InterpError)
               )
          )

    and evalExpExMatches env ex matches =
          (case foldl (evalExpMatchesF env ex) NONE matches of
             SOME resValue =>
               resValue
           | NONE =>
               (* Match-error - we raise the exception again. *)
               raise KsiVals.Pack ex
          )

    and evalExBindsF ({tag, ty, ann}, env) =
          (* We allocate a new unique "exception name" with ref () *)
          ST.insert env (ST.symbol tag)
                        (KsiVals.Exception {name = ref (), value = NONE})

    and evalExBinds env binds =
          foldl evalExBindsF env binds

  end
