(*
   This file is part of Anoq's implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003 Anoq of the Sun (alias Johnny Andersen)

   E-mail: anoq at HardcoreProcessing dot com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information. E-mail:
   info at HardcoreProcessing dot com

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiPrettyPrint =
  struct
    (* This module prints xi-Calculus types and restrictions using CeXL syntax. *)
    (* This should really be rewritten to use proper
       combinator pretty printing similar to what is found in
       the book "ML for the Working Programmer" by L.C. Paulson.
       But it is hand-coded here to keep the code simple. *)
    fun labsToStrs (lab::labs) acc =
          ", "::lab::(labsToStrs labs acc)
      | labsToStrs nil acc =
          acc

    fun labsToStr labs =
          String.concat (labsToStrs labs nil)

    fun restrToStrs KsiAbs.RstrNone acc =
          acc
      | restrToStrs (KsiAbs.RstrExclude {labs = nil}) acc =
          "{{}}"::acc
      | restrToStrs (KsiAbs.RstrExclude {labs = lab::labs}) acc =
          "{{"::lab::(labsToStrs labs ("}}"::acc))
      | restrToStrs (KsiAbs.RstrTyPats {pats = pat::pats}) acc =
          restrTyPatToStrs pat (restrTyPatsToStrs pats acc)
      | restrToStrs _ acc =
          (* A RstrTyPats without patterns would be an error in xi-Calculus.
             Our job here is only to display though - even errors. *)
          acc

    and restrTyPatsToStrs (pat::pats) acc =
          " / "::(restrTyPatToStrs pat (restrTyPatsToStrs pats acc))
      | restrTyPatsToStrs nil acc =
          acc

    and restrTyPatToStrs (KsiAbs.RstrTyPat {args = arg::args, name}) acc =
          "["::(restrTyPatArgToStrs arg
                  (restrTyPatArgsToStrs args ("] "::name::acc)))
      | restrTyPatToStrs (KsiAbs.RstrTyPat {args = nil, name}) acc =
          "[] "::name::acc

    and restrTyPatArgToStrs {name, pat} acc =
          name::" = "::(restrTyPatToStrs pat acc)

    and restrTyPatArgsToStrs (arg::args) acc =
          ", "::(restrTyPatArgToStrs arg (restrTyPatArgsToStrs args acc))
      | restrTyPatArgsToStrs nil acc =
          acc

    fun restrToStr ty =
          String.concat (restrToStrs ty nil)

    fun tyToStrs (KsiAbs.TyCon {args = arg::args, name, tags, ann}) acc =
          "["::(tyTyConArgToStrs arg
                  (tyTyConArgsToStrs args ("] "::name::acc)))
      | tyToStrs (KsiAbs.TyCon {args = nil, name, tags, ann}) acc =
          (* No tycon parameters - so we just display the name *)
          name::acc
      | tyToStrs (KsiAbs.TyFun {from, to, ann}) acc =
          tyToStrs from (" -> "::(tyToStrs to acc))
      | tyToStrs (KsiAbs.TyField {present, ty, ann}) acc =
          "("::(tyToStrs present (", "::(tyToStrs ty (") ?"::acc))))
      | tyToStrs (KsiAbs.TyExtRecord {fields = field::fields, ext, ann}) acc =
          "{"::(tyRecordFieldToStrs field
                  (tyRecordFieldsToStrs fields
                     (tyRecordExtToStrs ext ("}"::acc))))
      | tyToStrs (KsiAbs.TyExtRecord {fields = nil, ext, ann}) acc =
          (* We cannot display this type correctly if there are no
             fields at all. This situation is checked for elsewhere. *)
          "{<internal error: ext record with no fields>"::
          (tyRecordExtToStrs ext ("}"::acc))
      | tyToStrs (KsiAbs.TyEmptyRecord {ann}) acc =
          "{}"::acc
      | tyToStrs (KsiAbs.TyVar {name, restr, ann}) acc =
          name::acc
      | tyToStrs (KsiAbs.TyMeta {name, restr, bound, ann}) acc =
          (case !bound of
             NONE =>
               name::acc
           | SOME t =>
               tyToStrs t acc
          )

    and tyTyConArgsToStrs (arg::args) acc =
          ", "::(tyTyConArgToStrs arg (tyTyConArgsToStrs args acc))
      | tyTyConArgsToStrs nil acc =
          acc

    and tyTyConArgToStrs {name, ty} acc =
          name::" = "::(tyToStrs ty acc)

    and tyRecordFieldsToStrs (field::fields) acc =
          tyRecordFieldToStrs field (tyRecordFieldsToStrs fields acc)
      | tyRecordFieldsToStrs nil acc =
          acc

    and tyRecordFieldToStrs {lab, ty} acc =
          ", "::lab::" : "::(tyToStrs ty acc)

    and tyRecordExtToStrs (KsiAbs.TyEmptyRecord _) acc =
          (* Non-extensible record - no extension part to display *)
          acc
      | tyRecordExtToStrs (KsiAbs.TyExtRecord {fields, ext, ann}) acc =
          (* The record is longer still - continue to display *)
          tyRecordFieldsToStrs fields (tyRecordExtToStrs ext acc)
      | tyRecordExtToStrs (KsiAbs.TyMeta {name, restr, bound, ann}) acc =
          (case !bound of
             NONE =>
               (* Extension is an unbound meta variable *)
               ", ... "::name::acc
           | SOME t =>
               (* The record is longer still - continue to display *)
               tyRecordExtToStrs t acc
          )
      | tyRecordExtToStrs (KsiAbs.TyVar {name, restr, ann}) acc =
          (* Extension is a type variable *)
          ", ... "::name::acc
      | tyRecordExtToStrs t acc =
          (* Any other type (i.e. TyCon, TyFun or TyField)
             would be an error here - but we just display it. *)
          ", ... "::(tyToStrs t acc)
    
    fun tyToStr ty =
          String.concat (tyToStrs ty nil)
  end