(*
   This file is part of Anoq's implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003 Anoq of the Sun (alias Johnny Andersen)

   E-mail: anoq at HardcoreProcessing dot com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information. E-mail:
   info at HardcoreProcessing dot com

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiTypeInfer =
  struct
    exception InternalError

    structure ST = SymbolTable

    (* Unique names *)
    val uniqueMetaNames = CountingInteger.create ()

    (* Functions for reporting errors *)
    val errorReported = ref false

    fun reset () =
          (errorReported := false;
           CountingInteger.reset uniqueMetaNames)

    fun reportError str =
          (errorReported := true;
           print ("Type error: " ^ str ^ "\n"))

    val reportErrorStrs = reportError o String.concat

    fun reportInternalError str =
          (reportError str;
           raise InternalError)

    fun reportInternalErrorStrs strs =
          (reportErrorStrs strs;
           raise InternalError)

    fun newMetaName () =
          "'meta#" ^ (Int.toString (CountingInteger.next uniqueMetaNames))

    fun newMetaTyRestr restr =
          KsiAbs.TyMeta
            {name = newMetaName (),
             restr = restr,
             bound = ref NONE,
             ann = ()}

    fun newMetaTy () =
          newMetaTyRestr KsiAbs.RstrNone

    fun newAbsentField () =
          KsiAbs.TyField
            {present = KsiAbs.TyCon
                         {args = nil,
                          name = "absent",
                          tags = ["absent"],
                          ann = ()},
             ty = newMetaTy (),
             ann = ()}

    (* General utilities *)
    (* Test set membership *)
    fun hasElem eqF elem s =
          List.exists (fn e => eqF(elem, e)) s

    (* Test subset *)
    fun isSubset eqF subS superS =
          foldl (fn (e, acc) => acc andalso hasElem eqF e superS) true subS

    (* Test sets disjoint *)
    fun isDisjointSets eqF s1 s2 =
          foldl (fn (e, acc) => acc andalso not (hasElem eqF e s2)) true s1

    (* Test set equality *)
    fun isEqualSets eqF s1 s2 =
          isSubset eqF s1 s2 andalso
          isSubset eqF s2 s2

    (* Make set intersection *)
    fun makeIntersectSet eqF s1 s2 =
          List.filter (fn e => hasElem eqF e s1) s2

    (* Make set difference *)
    fun makeDifferenceSet eqF s removeS =
          List.filter (fn e => not (hasElem eqF e removeS)) s

    (* Make set union *)
    fun makeUnionSet eqF s1 s2 =
          (makeDifferenceSet eqF s1 s2) @ s2

    (* Equality tests *)
    fun eqTyPat (KsiAbs.RstrTyPat {args, name},
                 KsiAbs.RstrTyPat {args = args2, name = name2}) =
          name = name2 andalso
          isEqualSets eqTyPatArg args args2

    and eqTyPatArg ({name, pat}, {name = name2, pat = pat2}) =
          name = name2 andalso
          eqTyPat (pat, pat2)

    fun eqRecordField ({lab, ty}, {lab = lab2, ty = _}) =
          lab = lab2

    (* Equality predicates *)
    fun isNamedMetaVar theName (KsiAbs.TyMeta {name, ...}) =
          theName = name
      | isNamedMetaVar _ _ =
          false

    (* Traverse *)
    fun foldEmbedTy f acc (t as (KsiAbs.TyCon {args, name, tags, ann})) =
          f (t, foldEmbedTyTyConArgs f acc args)
      | foldEmbedTy f acc (t as (KsiAbs.TyFun {from, to, ann})) =
          f (t, foldEmbedTy f (foldEmbedTy f acc from) to)
      | foldEmbedTy f acc (t as (KsiAbs.TyField {present, ty, ann})) =
          f (t, foldEmbedTy f (foldEmbedTy f acc present) ty)
      | foldEmbedTy f acc (t as (KsiAbs.TyExtRecord {fields, ext, ann})) =
          f (t, foldEmbedTyExtRecordFields f acc fields)
      | foldEmbedTy f acc (t as (KsiAbs.TyEmptyRecord _)) =
          f (t, acc)
      | foldEmbedTy f acc (t as (KsiAbs.TyVar _)) =
          f (t, acc)
      | foldEmbedTy f acc (t as (KsiAbs.TyMeta {name, restr, bound, ann})) =
          f (t,
             case !bound of
               NONE    => acc
             | SOME ty => foldEmbedTy f acc ty
            )

    (* If we had extensible records here, it would be enough to write
       2 functions here rather than 4! We only use the field ty :-) *)
    and foldEmbedTyTyConArgs f acc args =
          foldl (foldEmbedTyTyConArg f) acc args

    and foldEmbedTyTyConArg f ({name, ty}, acc) =
          foldEmbedTy f acc ty

    and foldEmbedTyExtRecordFields f acc args =
          foldl (foldEmbedTyExtRecordField f) acc args

    and foldEmbedTyExtRecordField f ({lab, ty}, acc) =
          foldEmbedTy f acc ty

    (* Occurs check *)
    fun occursNamedMetaVar theName ty =
          foldEmbedTy (fn (t, acc) => acc orelse isNamedMetaVar theName t)
                      false ty

    (* Lookup function for order rows *)
    fun getOrderRowField r theLab =
          case List.find (fn {lab, ty} => theLab = lab) r of
            NONE =>
              reportInternalErrorStrs
                ["Internal error: Order row label ", theLab, " not found!"]
          | SOME {lab, ty} =>
              ty

    (* Lookup function for TyCon arguments *)
    fun getTyConArg args theName =
          case List.find (fn {name, ty} => theName = name) args of
            NONE =>
              reportInternalErrorStrs
                ["Internal error: TyCon argument name ", theName, " not found!"]
          | SOME {name, ty} =>
              ty

    (* Lookup function for TyPat arguments *)
    fun getTyPatArg patArgs theName =
          case List.find (fn {name, pat} => theName = name) patArgs of
            NONE =>
              reportInternalErrorStrs
                ["Internal error: TyPat argument name ", theName, " not found!"]
          | SOME {name, pat} =>
              pat

    (* Unification *)
    fun moreRestrictive (_, KsiAbs.RstrNone) =
          true
      | moreRestrictive (KsiAbs.RstrExclude {labs},
                         KsiAbs.RstrExclude {labs = labs2}) =
          (* labs2 must be a subset of labs *)
          isSubset op= labs2 labs
      | moreRestrictive (KsiAbs.RstrTyPats {pats},
                         KsiAbs.RstrTyPats {pats = pats2}) =
          let
            val len = length pats
            val len2 = length pats2
          in
            if 1 <= len andalso 1 <= len2 then
              len <= len2 andalso isSubset eqTyPat pats pats2
            else
              (reportError "Type pattern sequence pats may not be empty!";
               false)
          end
      | moreRestrictive (r1, r2) =
          (* r1 is not more restrictive than r2 *)
          false

    fun tyRespectsRestr (KsiAbs.TyVar {name, restr, ann}, theRestr) =
          moreRestrictive (restr, theRestr)
      | tyRespectsRestr (KsiAbs.TyMeta {name, restr, bound, ann}, theRestr) =
          moreRestrictive (restr, theRestr)
      | tyRespectsRestr (t, KsiAbs.RstrNone) =
          true
      | tyRespectsRestr (t, KsiAbs.RstrTyPats {pats}) =
          List.exists (fn pat => tyRespectsRestrTyPat (t, pat)) pats
      | tyRespectsRestr (KsiAbs.TyExtRecord {fields, ext, ann}, theRestr) =
          (case theRestr of
             KsiAbs.RstrExclude {labs} =>
               let
                 val labs2 = map #lab fields
               in
                 if isDisjointSets op= labs labs2 then
                   tyRespectsRestr
                     (ext, KsiAbs.RstrExclude {labs = labs @ labs2})
                 else
                   (* The record labes were not disjoint *)
                   false
               end
           | _ =>
               false
          )
      | tyRespectsRestr (KsiAbs.TyEmptyRecord {ann}, theRestr) =
          (case theRestr of
             KsiAbs.RstrExclude _ =>
               true
           | _ =>
               false
          )
      | tyRespectsRestr _ =
          (* Anything else does not respect the restriction *)
          false

    and tyRespectsRestrTyPat (KsiAbs.TyMeta {name, restr, bound, ann}, thePat) =
          (case !bound of
             NONE =>
               (* We treat this as an error in the semantics.
                  Otherwise it would require unification of restrictions
                  which is NOT trivial to add to xi-Calculus! *)
               false
           | SOME t =>
               tyRespectsRestrTyPat (t, thePat)
          )
      | tyRespectsRestrTyPat (KsiAbs.TyVar {name, restr, ann}, thePat) =
          moreRestrictive (restr, KsiAbs.RstrTyPats {pats = [thePat]})
      | tyRespectsRestrTyPat (KsiAbs.TyCon {args, name, tags, ann = _},
                              KsiAbs.RstrTyPat {args = patArgs, name = patName}) =
          if name = patName then
            (* The type names match - check the parameters *)
            let
              val argNames = map #name args
              val patArgNames = map #name patArgs
            in
              if isEqualSets op= argNames patArgNames then
                tyRespectsRestrTyConArgs args patArgs
              else
                (* If args and patArgs does not have the same parameters
                   it must be an error in the parser or the translation to
                   xi-Calculus! *)
                reportInternalErrorStrs
                  ["Internal error: The constructor ", name,
                   " occurs with differing sets of parameter names ",
                   "in a restriction!"]
            end
          else
            (* Type names do not match *)
            false
      | tyRespectsRestrTyPat (_, _) =
          (* Anything else is irrespectful of the type :-) *)
          false
      
    and tyRespectsRestrTyConArgs args patArgs =
          foldl (fn ({name, ty}, acc) =>
                   acc andalso
                   (tyRespectsRestrTyPat (ty, getTyPatArg patArgs name)))
                true args

    (* Make record fields independent of order *)
    fun makeOrderRow (ty as (KsiAbs.TyExtRecord {fields, ext, ann})) exclLabs =
          (case fields of
             nil =>
               (reportError "Encountered extensible record without any fields!";
                makeOrderRow ext exclLabs)
           | _ =>
               (* There is at least one field as required *)
               let
                 val labs = map #lab fields
               in
                 if isDisjointSets op= labs exclLabs then
                   let
                     val (orderRow, ty) = makeOrderRow ext (labs @ exclLabs)
                   in
                     (fields @ orderRow, ty)
                   end
                 else
                   let
                     val forbiddenLabs = makeIntersectSet op= labs exclLabs
                   in
                     (reportErrorStrs
                        ("Record contains the following forbidden fields: "::
                         (KsiPrettyPrint.labsToStrs forbiddenLabs nil));
                      makeOrderRow ext (labs @ exclLabs))
                   end
               end
          )
      | makeOrderRow (ty as (KsiAbs.TyEmptyRecord _)) exclLabs =
          (nil, ty)
      | makeOrderRow (ty as (KsiAbs.TyMeta {name, restr, bound, ann})) exclLabs =
          (case !bound of
             NONE =>
               (case restr of
                  KsiAbs.RstrExclude {labs} =>
                    if isSubset op= exclLabs labs then
                      (nil, ty)
                    else
                      let
                        val freshLabs = makeUnionSet op= exclLabs labs
                        val freshTy = newMetaTyRestr
                                        (KsiAbs.RstrExclude
                                           {labs = freshLabs})
                        val _ = bound := SOME freshTy
                      in
                        (nil, freshTy)
                      end
                | KsiAbs.RstrNone =>
                    let
                      val freshTy = newMetaTyRestr
                                      (KsiAbs.RstrExclude
                                         {labs = exclLabs})
                      val _ = bound := SOME freshTy
                    in
                      (nil, freshTy)
                    end
                | _ =>
                    (reportErrorStrs
                       ["Record does not unify with the meta variable: ", name,
                        " since it is restricted to be a non-record type"];
                     (nil, ty))
               )
           | SOME t =>
               makeOrderRow t exclLabs
          )
      | makeOrderRow (ty as (KsiAbs.TyVar {name, restr, ann})) exclLabs =
          (case restr of
             KsiAbs.RstrExclude {labs} =>
               if isSubset op= exclLabs labs then
                 (nil, ty)
               else
                 let
                   val forbiddenLabs = makeDifferenceSet op= exclLabs labs
                 in
                   (reportErrorStrs
                      ("Record does not unify with the type variable: "::name::
                       " since it contains the following forbidden fields: "::
                       (KsiPrettyPrint.labsToStrs forbiddenLabs nil));
                    (nil, ty))
                 end
           | _ =>
               (reportErrorStrs
                  ["Record does not unify with the type variable: ", name,
                   " since it is not restricted to be a record type"];
                (nil, ty))
          )
      | makeOrderRow ty exclLabs =
          (reportErrorStrs
             ["The type: ", KsiPrettyPrint.tyToStr ty,
              " is not a record where this is expected"];
           (nil, ty))
    
    fun reportTooManyLabelsDifferences (nil, nil) =
          (* Hey! The sets were supposed to be different! *)
          reportInternalError
            "Internal error - most likely in the set-operations!"
      | reportTooManyLabelsDifferences (nil, tooMany2) =
          reportErrorStrs
            ["Two non-extensible records failed to unify ",
             "because the second record contains the following labels ",
             "which are not present in the first record:\n",
             KsiPrettyPrint.labsToStr tooMany2]
      | reportTooManyLabelsDifferences (tooMany, nil) =
          reportErrorStrs
            ["Two non-extensible records failed to unify ",
             "because the first record contains the following labels ",
             "which are not present in the second record:\n",
             KsiPrettyPrint.labsToStr tooMany]
      | reportTooManyLabelsDifferences (tooMany, tooMany2) =
          reportErrorStrs
            ["Two non-extensible records failed to unify ",
             "because the first record contains the following labels ",
             "which are not present in the second record:\n",
             KsiPrettyPrint.labsToStr tooMany,
             "\nand the second record contains the following labels ",
             "which are not present in the first record:\n",
             KsiPrettyPrint.labsToStr tooMany2]

    fun reportOrderRowDifferences (labs, labs2) =
          let
            val tooMany = makeDifferenceSet op= labs labs2
            val tooMany2 = makeDifferenceSet op= labs2 labs
          in
            reportTooManyLabelsDifferences (tooMany, tooMany2)
          end

    fun unify (t1 as (KsiAbs.TyMeta {name, restr, bound, ann}) : unit KsiAbs.Ty, t2 : unit KsiAbs.Ty) =
          (case !bound of
             SOME t =>
               (unify (t, t2)) : unit
           | NONE =>
               (case t2 of
                  (KsiAbs.TyMeta
                     {name = name2, restr = restr2,
                      bound = bound2, ann}) =>
                     (case !bound2 of
                        SOME t =>
                          (* Since the mete variable name2 is bound to
                             a type, we have that name2 <> name,
                             since meta var name is not bound to any type. *)
                          unify (t1, t)
                      | NONE =>
                          if name = name2 then
                            (* This is OK *)
                            ()
                          else
                            if moreRestrictive (restr2, restr) then
                              bound := (SOME t2)
                            else if moreRestrictive (restr, restr2) then
                              bound2 := (SOME t1)
                            else
                              reportErrorStrs
                                ["Incompatiple restrictions: ",
                                 KsiPrettyPrint.restrToStr restr,
                                 " and: ",
                                 KsiPrettyPrint.restrToStr restr2]
                     )
                | _ =>
                    if occursNamedMetaVar name t2 then
                      reportErrorStrs [name, " occurs in t2!"]
                    else
                      if tyRespectsRestr (t2, restr) then
                        bound := (SOME t2)
                      else
                        reportErrorStrs
                          ["The type: ", KsiPrettyPrint.tyToStr t2,
                           " does not respect the restriction: ",
                           KsiPrettyPrint.restrToStr restr]
               )
          )
      | unify (t1, t2 as (KsiAbs.TyMeta _)) =
          (* Unification is symmetric *)
          unify (t2, t1)
      | unify (KsiAbs.TyVar {name, restr, ann},
               KsiAbs.TyVar {name = name2, restr = restr2, ann = _}) =
          if name = name2 then
            (* This is OK *)
            ()
          else
            reportErrorStrs
              ["Type variables ", name, " and ", name2, " cannot be unified!"]
      | unify (KsiAbs.TyFun {from, to, ann},
               KsiAbs.TyFun {from = from2, to = to2, ann = _}) =
          (unify (from, from2);
           unify (to, to2))
      | unify (KsiAbs.TyField {present, ty, ann},
               KsiAbs.TyField {present = present2, ty = ty2, ann = _}) =
          (unify (present, present2);
           unify (ty, ty2))
      | unify (KsiAbs.TyCon {args, name, tags, ann},
               KsiAbs.TyCon {args = args2, name = name2, tags = tags2, ann = _}) =
          if name = name2 then
            let
              val argNames = map #name args
              val argNames2 = map #name args2
            in
              if isEqualSets op= argNames argNames2 then
                unifyTyConArgs args args2
              else
                (* If args and args2 does not have the same parameters
                   it must be an error in the parser or the translation to
                   xi-Calculus! *)
                reportInternalErrorStrs
                  ["Internal error: The constructor ", name,
                   " occurs 2 places with differing sets of parameter names!"]
            end
          else
            reportErrorStrs
              ["Differing type constructors ", name, " and ",
               name2, " cannot be unified!"]
      | unify (t as (KsiAbs.TyExtRecord {fields, ext, ann}),
               t2 as (KsiAbs.TyExtRecord {fields = fields2, ext = ext2, ann = _})) =
          (* The syntactic restrictions must ensure that
             fields and fields2 are not nil.
             It is also double-checked by makeOrderRow. *)
          unifyOrderRow (makeOrderRow t nil)  (makeOrderRow t2 nil)
      | unify (t as (KsiAbs.TyExtRecord {fields, ext, ann}),
               t2 as (KsiAbs.TyEmptyRecord _)) =
          (* The syntactic restrictions must ensure that fields is not nil.
             It is also double-checked by makeOrderRow. *)
          unifyOrderRowWithEmpty (makeOrderRow t nil)
      | unify (t as (KsiAbs.TyEmptyRecord _),
               t2 as (KsiAbs.TyExtRecord {fields = fields2, ext = ext2, ann = _})) =
          (* The syntactic restrictions must ensure that fields is not nil.
             It is also double-checked by makeOrderRow. *)
          unifyOrderRowWithEmpty (makeOrderRow t2 nil)
      | unify (KsiAbs.TyEmptyRecord _, KsiAbs.TyEmptyRecord _) =
          (* This is OK *)
          ()
      | unify (t, t2) =
          reportErrorStrs
            ["Cannot unify the type:\n",
             KsiPrettyPrint.tyToStr t,
             "\nwith the type:\n",
             KsiPrettyPrint.tyToStr t2]

    and unifyTyConArgs args args2 =
          List.app (fn {name, ty} => unify (ty, getTyConArg args2 name)) args

    and unifyOrderRow ((r,  KsiAbs.TyEmptyRecord _) : ((unit KsiAbs.OrderRow) * (unit KsiAbs.Ty)))
                      ((r2, KsiAbs.TyEmptyRecord _) : ((unit KsiAbs.OrderRow) * (unit KsiAbs.Ty))) =
          let
            val labs = map #lab r
            val labs2 = map #lab r2
          in
            if isEqualSets op= labs labs2 then
              (* The record domains are equal as required *)
              (unifyOrderRowFields (r : unit KsiAbs.OrderRow)
                                   (r2 : unit KsiAbs.OrderRow)) : unit
            else
              (reportOrderRowDifferences (labs, labs2)) : unit
          end
      | unifyOrderRow (r,  KsiAbs.TyVar {name, restr, ann})
                      (r2, KsiAbs.TyVar {name = name2, restr = restr2, ann = _}) =
          if name = name2 then
            (* The type variables are the same as required *)
            (* Since the type variables comes from an order row
               it is already guaranteed that the restrictions are
               excluding labels and that they are OK. *)
            let
              val labs = map #lab r
              val labs2 = map #lab r2
            in
              if isEqualSets op= labs labs2 then
                (* The record domains are equal as required *)
                unifyOrderRowFields r r2
              else
                reportOrderRowDifferences (labs, labs2)
            end
          else
            reportErrorStrs
              ["Differing type variables: ", name, " and: ", name2,
               " cannot be unified"]
      | unifyOrderRow (r,  KsiAbs.TyMeta {name, restr, bound, ann})
                      (r2, KsiAbs.TyMeta {name = name2, restr = restr2,
                                           bound = bound2, ann = ann2}) =
          (case (restr, restr2) of
             (KsiAbs.RstrExclude {labs = labs},
              KsiAbs.RstrExclude {labs = labs2}) =>
                (* Restrictions are records as required *)
                let
                  val rInters = makeIntersectSet eqRecordField r r2
                  val rExtra = makeDifferenceSet eqRecordField r r2
                  val rExtra2 = makeDifferenceSet eqRecordField r2 r
                  val labsExtra = map #lab rExtra
                  val labsExtra2 = map #lab rExtra2
                in
                  if isDisjointSets op= labsExtra labs2 then
                    if isDisjointSets op= labsExtra2 labs then
                      (case (!bound, !bound2) of
                         (NONE, NONE) =>
                            let
                              (* First we handle meta variables
                                 since it is a side effect *)
                              val freshLabs = makeUnionSet op= labs labs2
                              val freshTy = newMetaTyRestr
                                              (KsiAbs.RstrExclude
                                                 {labs = freshLabs})
                              val rest = KsiAbs.TyExtRecord
                                            {fields = rExtra,
                                             ext = freshTy,
                                             ann = ann}
                              val rest2 = KsiAbs.TyExtRecord
                                            {fields = rExtra2,
                                             ext = freshTy,
                                             ann = ann2}
                              val _ = bound := SOME rest2
                              val _ = bound2 := SOME rest
                            in
                              (* Unify the record fields in the
                                 intersection of the 2 records *)
                              unifyOrderRowFields rInters r2
                            end
                       | _ =>
                            reportInternalErrorStrs
                              ["Internal error: A meta variable for an ",
                               "order row is bound"]
                      )
                    else
                      let
                        val forbiddenLabs2 = makeIntersectSet op= labsExtra2 labs
                      in
                        reportErrorStrs
                          ("Two records do not unify because one record "::
                           "contains the following forbidden fields: "::
                           (KsiPrettyPrint.labsToStrs forbiddenLabs2 nil))
                      end
                  else
                    let
                      val forbiddenLabs = makeIntersectSet op= labsExtra labs2
                    in
                      reportErrorStrs
                        ("Two records do not unify because one record "::
                         "contains the following forbidden fields: "::
                         (KsiPrettyPrint.labsToStrs forbiddenLabs nil))
                    end
                end
           | _ =>
                reportInternalErrorStrs
                  ["Internal error: Meta variables for order row are not ",
                   "restricted to be records as expected"]
          )
      | unifyOrderRow (r,  KsiAbs.TyMeta {name, restr, bound, ann})
                      (r2, ty2 as (KsiAbs.TyVar {name = name2, restr = restr2, ann = ann2})) =
          (case (restr, restr2) of
             (KsiAbs.RstrExclude {labs = labs},
              KsiAbs.RstrExclude {labs = labs2}) =>
                (* Restrictions are records as required *)
                let
                  val rLabs = map #lab r
                  val rLabs2 = map #lab r2
                  val rExtra2 = makeDifferenceSet eqRecordField r2 r
                  val labsExtra2 = map #lab rExtra2
                in
                  if isDisjointSets op= labsExtra2 labs then
                    if isSubset op= rLabs rLabs2 then
                      if isSubset op= labs labs2 then
                        (case !bound of
                           NONE =>
                              let
                                (* First we handle meta variables
                                   since it is a side effect *)
                                val rest = KsiAbs.TyExtRecord
                                             {fields = rExtra2,
                                              ext = ty2,
                                              ann = ann2}
                                val _ = bound := SOME rest
                              in
                                (* Unify the record fields in the record r *)
                                unifyOrderRowFields r r2
                              end
                         | _ =>
                              reportInternalErrorStrs
                                ["Internal error: A meta variable for an ",
                                 "order row is bound"]
                        )
                      else
                        let
                          val forbiddenLabs = makeDifferenceSet op= labs labs2
                        in
                          reportErrorStrs
                            ("Two records do not unify because one record "::
                             "is extended with a meta variable restricted "::
                             "with the following extra fields: "::
                             (KsiPrettyPrint.labsToStrs forbiddenLabs nil))
                      end
                    else
                      let
                        val forbiddenLabs = makeDifferenceSet op= rLabs rLabs2
                      in
                        reportErrorStrs
                          ("Two records do not unify because one record "::
                           "contains the following forbidden fields: "::
                           (KsiPrettyPrint.labsToStrs forbiddenLabs nil))
                      end
                  else
                    let
                      val forbiddenLabs2 = makeIntersectSet op= labsExtra2 labs
                    in
                      reportErrorStrs
                        ("Two records do not unify because one record "::
                         "contains the following forbidden fields: "::
                         (KsiPrettyPrint.labsToStrs forbiddenLabs2 nil))
                    end
                end
           | _ =>
                reportInternalErrorStrs
                  ["Internal error: Meta variable or type variable for ",
                   "order row are not restricted to be records as expected"]
          )
      | unifyOrderRow (r,  KsiAbs.TyMeta {name, restr, bound, ann})
                      (r2, ty2 as (KsiAbs.TyEmptyRecord {ann = ann2})) =
          (case restr of
             KsiAbs.RstrExclude {labs} =>
                (* Restrictions are records as required *)
                let
                  val rInters = makeIntersectSet eqRecordField r r2
                  val rExtra = makeDifferenceSet eqRecordField r r2
                  val rExtra2 = makeDifferenceSet eqRecordField r2 r
                  val labsExtra2 = map #lab rExtra2
                in
                  if isDisjointSets op= labsExtra2 labs then
                    (case !bound of
                       NONE =>
                         let
                           (* First we handle meta variables
                              since it is a side effect *)
                           val rest = KsiAbs.TyExtRecord
                                        {fields = rExtra2,
                                         ext = ty2,
                                         ann = ann2}
                           val _ = bound := SOME rest

                           (* Unify the record fields which are in r
                              but not in r2 with absent fields. *)
                           val _ = unifyAbsentFields rExtra
                         in
                           (* Unify the record fields in the
                              intersection of the 2 records *)
                           unifyOrderRowFields rInters r2
                         end
                     | _ =>
                         reportInternalErrorStrs
                           ["Internal error: A meta variable for an ",
                            "order row is bound"]
                    )
                  else
                    let
                      val forbiddenLabs2 = makeIntersectSet op= labsExtra2 labs
                    in
                      reportErrorStrs
                        ("Two records do not unify because one record "::
                         "contains the following forbidden fields: "::
                         (KsiPrettyPrint.labsToStrs forbiddenLabs2 nil))
                    end
                end
           | _ =>
                reportInternalErrorStrs
                  ["Internal error: Meta variable for ",
                   "order row is not restricted to be a record as expected"]
          )
      | unifyOrderRow (r,  ty  as (KsiAbs.TyVar _))
                      (r2, ty2 as (KsiAbs.TyMeta _)) =
          (* Unification is symmetric *)
          unifyOrderRow (r2, ty2) (r, ty)
      | unifyOrderRow (r,  ty  as (KsiAbs.TyEmptyRecord _))
                      (r2, ty2 as (KsiAbs.TyMeta _)) =
          (* Unification is symmetric *)
          unifyOrderRow (r2, ty2) (r, ty)
      | unifyOrderRow _ _ =
          reportInternalError
            "Internal error: Two order rows have wrong types"

    and unifyOrderRowFields r r2 =
          List.app (fn {lab, ty} => unify (ty, getOrderRowField r2 lab)) r
    
    and unifyAbsentFields r =
          List.app (fn {lab, ty} => unify (ty, newAbsentField ())) r

    and unifyOrderRowWithEmpty (r, KsiAbs.TyEmptyRecord _) =
          unifyAbsentFields r
      | unifyOrderRowWithEmpty (r, KsiAbs.TyMeta {name, restr, bound, ann}) =
          (case !bound of
             NONE =>
               (* First bind type variable since this causes a side-effect. *)
               (bound := SOME (KsiAbs.TyEmptyRecord {ann = ()});
                unifyAbsentFields r)
           | _ =>
               reportInternalErrorStrs
                 ["Internal error: A meta variable for an ",
                  "order row is bound"]
          )
      | unifyOrderRowWithEmpty (r, ty) =
          reportErrorStrs
            ["Cannot unify the type: ",
             KsiPrettyPrint.tyToStr ty,
             " with the empty record"]

    (* Inference rules for xi-Calculus unification *)
    fun transExp (vE, tyvE) (KsiAbs.ExpFun {param, exp, ann}) =
          let
            val paramTy = newMetaTy ()
            val vE' = ST.insert vE (ST.symbol param) paramTy
            val (exp', toTy) = transExp (vE', tyvE) exp
            val annTy = KsiAbs.TyFun {from = paramTy, to = toTy, ann = ()}
          in
            (KsiAbs.ExpFun {param = param, exp = exp', ann = annTy},
             annTy)
          end
      | transExp (vE, tyvE) (KsiAbs.ExpApp {funExp, argExp, ann}) =
          let
            val (argExp', argTy) =  transExp (vE, tyvE) argExp
            val (funExp', funTy) =  transExp (vE, tyvE) funExp
            
            val annTy = newMetaTy ()
            val _ = unify (funTy, KsiAbs.TyFun
                                    {from = argTy,
                                     to = annTy,
                                     ann = ()})
          in
            (KsiAbs.ExpApp {funExp = funExp', argExp = argExp', ann = annTy},
             annTy)
          end

      | transExp (vE, tyvE) (exp as (KsiAbs.ExpEmptyRecord {ann})) =
          transExpRec (vE, tyvE, nil) exp
      | transExp (vE, tyvE) (exp as (KsiAbs.ExpExtRecord {fields, ext, ann})) =
          transExpRec (vE, tyvE, nil) exp

    and transExpRec (vE, tynE, exclLabs) (KsiAbs.ExpEmptyRecord {ann}) =
          let
            val annTy = KsiAbs.TyEmptyRecord {ann = ()}
          in
            (KsiAbs.ExpEmptyRecord {ann = annTy},
             annTy)
          end
    (*
      | transExpRec (vE, tynE, exclLabs) (KsiAbs.ExpExtRecord {fields, ext, ann}) =
          let
          in
            (KsiAbs.ExpExtRecord ,
             annTy)
          end
    *)
  end
